#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright 2020 by leodflag

"""This module contains statistics calculation function.

Addition, subtraction, multiplication, and division of lists, as well as basic statistical calculation functions 
such as variance, standard deviation, and autocorrelation coefficient.

Functions
-----------
chack_list_all_num(listA) -> bool
	Determine the content of the list as one numerical data.

plus(listA, listB) -> list
	The two lists are added element by element.

sub(listA, listB) -> list
	Subtract two lists element by element.

mult(listA, listB) -> list
	Multiply two lists element by element.

div(listA, listB) -> list
	Divide two lists element by element.

log(listA) -> list:
    Multiply all the numerical elements in the list by the natural logarithm.

round_list(listA, ndigits: int) -> list:
    Take all the values in the list to a few digits after the decimal point.

mean(listA) -> float
	Calculate arithmetic mean of numerical data in a list.

variance(listA) -> float
	Calculate the variance of all elements in a list.

standard_deviation(listA) -> float
	Calculate the standard deviation of all elements in a list.

covariance(listA, listB) -> float
	Calculate the covariance of two lists of one-dimensional data.

coefficient_of_variation(listA) -> float
	Calculate the coefficient of variation of all elements in a list.

correlation_coefficient(listA, listB) -> float
	Calculate the linear correlation coefficient  of two lists of one-dimensional data.

coefficient_of_determination(listA, listB) -> float
	Calculate the coefficient of determination  of two lists of one-dimensional data.

mean_square_error(target_list, predict_list) -> float
	The regression index used to calculate the forecast error. Calculate the mean square error  of two lists of one-dimensional data.

mean_absolute_error(target_list, predict_list) -> float:
    The regression index used to calculate the forecast error. Calculate the mean absolute error(MAE) of two lists of one-dimensional data.

root_mean_squard_error(target_list, predict_list) -> float:
    The regression index used to calculate the forecast error. Calculate the root mean square error(RMSE) of two lists of one-dimensional data.

normalized_mean_squard_error(target_list, predict_list) -> float:
    The regression index used to calculate the forecast error. Calculate the normalized mean square error(NRMSE) of two lists of one-dimensional data.

ACF(listA, listB) -> float
	Calculate the autocorrelation coefficients of two lists of one-dimensional data.

Notes
-----------
Pay attention to the input data type of the function.

Use the function "list_to_dataframe" or "dataframe_to_list" of the module "TimeSeriesAnalysis.ProcessData" to convert the data type.

"""
import math
import numpy as np
import pandas as pd

def chack_list_all_num(listA) -> bool:
    """Determine the content of the list as one numerical data.

    This function is to check the input list, the content is a one-dimensional numerical list.
    If the input is not a one-dimensional list, or the content of the one-dimensional list is not all numerical data, 
    it will be returned 'False', otherwise it will return 'True'.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    bool(True or False).

    'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, 
    and one-dimensional non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5].

    Error
    ---------
        ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
        Solution: The entered'listA' is an empty list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        TypeError: object of type 'NoneType' has no len()
        Solution:  Please do not enter 'None'.Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        if(len(listA) != 0):
            for i in range(len(listA)):
                if(type(listA[i]) == int or type(listA[i]) == float or type(listA[i]) == np.float64):
                    continue
                else:
                    return False
            return True
        else:
            raise ValueError("The list must be one-dimensional numerical data, and there is at least one numerical data in it.")
    except KeyError:
        raise KeyError("The'data' type is'pandas.DataFrame', please enter the correct key value.")
    except TypeError as err:
        raise TypeError(err)


def plus(listA, listB) -> list:
    """The two lists are added element by element.

    Addition: The two lists are added element by element, and return a list of sums.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    list. The result of adding the two lists.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        chack_A = chack_list_all_num(listA)
        chack_B = chack_list_all_num(listB)
        if(chack_A and chack_B):
            if (len(listA) == len(listB)):
                result_list = []
                for i in range(len(listA)):
                    result_list.append(listA[i] + listB[i])
                return result_list
            else:
                raise ValueError("The two lists must be the same length.")
        else:
            raise ValueError("List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B)
    except TypeError as err:
        raise TypeError(err)

def sub(listA, listB) -> list:
    """Subtract two lists element by element.

    Subtraction: Subtract the elements of two lists pair by pair, and return a list of results.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    list. The result of subtracting the two lists.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        chack_A = chack_list_all_num(listA)
        chack_B = chack_list_all_num(listB)
        if(chack_A and chack_B):
            if (len(listA) == len(listB)):
                result_list = []
                for i in range(len(listA)):
                    result_list.append(listA[i] - listB[i])
                return result_list
            else:
                raise ValueError("The two lists must be the same length.")
        else:
            raise ValueError("List content must be one numerical data: listA=",chack_A, "; listB=", chack_B)
    except TypeError as err:
        raise TypeError(err)

def mult(listA, listB) -> list:
    """Multiply two lists element by element.

    Multiplication: The elements of two lists are multiplied pair by pair to return a list of results.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    list. The result of multiplication.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        chack_A = chack_list_all_num(listA)
        chack_B = chack_list_all_num(listB)
        if(chack_A and chack_B):
            if (len(listA) == len(listB)):
                result_list = []
                for i in range(len(listA)):
                    result_list.append(listA[i] * listB[i])
                return result_list
            else:
                raise ValueError("The two lists must be the same length.")
        else:
            raise ValueError("List content must be one numerical data: listA=",chack_A, "; listB=", chack_B)
    except TypeError as err:
        raise TypeError(err)

def div(listA, listB) -> list:
    """Divide two lists element by element.

    Division: Divide the elements of two lists pair by pair, and return a list of results.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    list. Result of division.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        chack_A = chack_list_all_num(listA)
        chack_B = chack_list_all_num(listB)
        if(chack_A and chack_B):
            if (len(listA) == len(listB)):
                result_list = []
                for i in range(len(listA)):
                    result_list.append(listA[i] / listB[i])
                return result_list
            else:
                raise ValueError("The two lists must be the same length.")
        else:
            raise ValueError("List content must be one numerical data: listA=",chack_A, "; listB=", chack_B)
    except TypeError as err:
        raise TypeError(err)

def log(listA) -> list:
    """Multiply all the numerical elements in the list by the natural logarithm.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    list. Result of natural logarithm list.

    Error
    ---------
        ValueError: List content must be one-dimensional numerical data: listA=",chack_A".
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    log: https://en.wikipedia.org/wiki/Natural_logarithm

    """
    try:
        chack_A = chack_list_all_num(listA)
        if(chack_A):
            result_list = []
            for i in range(len(listA)):
                if (listA[i] > 0):
                    result_list.append(math.log(listA[i]))
                else:
                    result_list.append(0)
            return result_list
        else:
            raise ValueError("List content must be one numerical data: listA=",chack_A)
    except TypeError as err:
        raise TypeError(err)


def round_list(listA, ndigits: int) -> list:
    """Take all the values in the list to a few digits after the decimal point.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        ndigits: int
        A few digits after the decimal point.

    Returns
    ---------
    list. Return a list of values with a few digits after the decimal point.

    Error
    ---------
        ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
        Solution: The entered'listA' is an empty list. Please change to a one-dimensional list of numerical data.

        ValueError: The list must be one-dimensional numerical data.
        Solution: 'listA' contains a string or is two-dimensional list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        chack_A = chack_list_all_num(listA)
        if(chack_A):
            result_list = []
            for i in range(len(listA)):
                result_list.append(round(listA[i], ndigits))
            return result_list
        else:
            raise ValueError("List content must be one numerical data: listA=",chack_A)
    except TypeError as err:
        raise TypeError(err)

def mean(listA) -> float:
    """Calculate arithmetic mean of numerical data in a list.

    Mean: Add all the numerical data in the list and divide by the length of the list.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. Arithmetic mean of numerical data in a list.

    Error
    ---------
        ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
        Solution: The entered'listA' is an empty list. Please change to a one-dimensional list of numerical data.

        ValueError: The list must be one-dimensional numerical data.
        Solution: 'listA' contains a string or is two-dimensional list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Mean: https://zh.wikipedia.org/wiki/%E5%B9%B3%E5%9D%87%E6%95%B0

    """
    try:
        chack_A = chack_list_all_num(listA)
        if (chack_A):
            Sum = 0
            for i in range(len(listA)):
                Sum += listA[i]
            return Sum/(len(listA))
        else:
            raise ValueError("The list must be one-dimensional numerical data.")
    except TypeError as err:
        raise TypeError(err)

def variance(listA) -> float:
    """Calculate the variance of all elements in a list.

    Variance:The average of the squared distance of each number from its mean. 
    Algorithm: All the numerical data in the list are subtracted one by one from the mean value 
    and then squared, after adding them up and dividing by the length of the list minus one.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The variance of all elements in a list.

    Error
    ---------
        ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
        Solution: The entered'listA' is an empty list. Please change to a one-dimensional list of numerical data.

        ValueError: The list must be one-dimensional numerical data.
        Solution: 'listA' contains a string or is two-dimensional list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Variance: https://zh.wikipedia.org/wiki/%E6%96%B9%E5%B7%AE

    """
    try:
        listA_mean = mean(listA)
        Sum = 0
        for i in range(len(listA)):
            num = 0
            num = (listA[i] - listA_mean)**2
            Sum += num
        return Sum / (len(listA) - 1)
    except TypeError as err:
        raise TypeError(err)

def standard_deviation(listA) -> float:
    """Calculate the standard deviation of all elements in a list.

    Standard deviation: Square root of variance. All the numerical data in the list are subtracted one by one 
    from the mean value and then squared. Algorithm: After adding them up, divide by the length of the list minus 
    one, and finally take the square root.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The standard deviation of all elements in a list.

    Error
    ---------
        ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
        Solution: The entered'listA' is an empty list. Please change to a one-dimensional list of numerical data.

        ValueError: The list must be one-dimensional numerical data.
        Solution: 'listA' contains a string or is two-dimensional list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Standard deviation: https://zh.wikipedia.org/wiki/%E6%A8%99%E6%BA%96%E5%B7%AE

    """
    try:
        listA_mean = mean(listA)
        Sum = 0
        for i in range(len(listA)):
            num = 0
            num = (listA[i] - listA_mean)**2
            Sum += num
        return (Sum /( len(listA) - 1))**0.5
    except TypeError as err:
        raise TypeError(err)

def covariance(listA, listB) -> float:
    """Calculate the covariance of two lists of one-dimensional data.

    Covariance: After calculating the respective averages of the two lists, subtract the averages one by one and 
    multiply them to form a value. Algorithm: After the entire list is calculated, all the values are added up, and finally
    divided by the length of the list minus one.

    Parameters
    ---------
        listA: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The covariance of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Covariance: https://zh.wikipedia.org/wiki/%E5%8D%8F%E6%96%B9%E5%B7%AE

    """
    try:
        listA_mean = mean(listA)
        listB_mean = mean(listB)
        length = len(listA)
        Sum = 0
        for i in range(length):
            Sum += (listA[i] - listA_mean)*(listB[i] - listB_mean)
        return Sum / (length - 1)
    except TypeError as err:
        raise TypeError(err)

def coefficient_of_variation(listA) -> float:
    """Calculate the coefficient of variation of all elements in a list.

    Coefficient of variation: Can compare the standard deviation between different unit variables. 
    Algorithm: First calculate the mean and standard deviation of the list, and finally divide the standard deviation 
    by the mean and multiply by 100.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The coefficient of variation of all elements in a list.

    Error
    ---------
        ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
        Solution: The entered'listA' is an empty list. Please change to a one-dimensional list of numerical data.

        ValueError: The list must be one-dimensional numerical data.
        Solution: 'listA' contains a string or is two-dimensional list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Coefficient of variation: https://zh.wikipedia.org/wiki/%E5%8F%98%E5%BC%82%E7%B3%BB%E6%95%B0

    """
    try:
        listA_mean = mean(listA)
        listA_sd = standard_deviation(listA)
        return (listA_sd / listA_mean)*100
    except TypeError as err:
        raise TypeError(err)

def correlation_coefficient(listA, listB) -> float:
    """Calculate the linear correlation coefficient  of two lists of one-dimensional data.

    Correlation coefficient: The strength of linear correlation between two variables. 
    Algorithm: Calculate the linear correlation coefficient according to the formula.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The linear correlation coefficient  of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Correlation coefficient: https://www.statisticshowto.com/probability-and-statistics/correlation-coefficient-formula/

    """
    try:
        listAB_mult_sum = sum(mult(listA, listB))
        listAB_sum_mult = sum(listA) * sum(listB)
        listAA_mult_sum = sum(mult(listA, listA))
        listAA_sum_mult = sum(listA) * sum(listA)
        listBB_mult_sum = sum(mult(listB, listB))
        listBB_sum_mult = sum(listB) * sum(listB)
        n = len(listA)
        r =( (n*listAB_mult_sum - listAB_sum_mult) /
            (( (n*listAA_mult_sum - listAA_sum_mult)*(n*listBB_mult_sum - listBB_sum_mult) )**0.5) )
        return r
    except TypeError as err:
        raise TypeError(err)

def coefficient_of_determination(train_target, target_list, predict_list) -> float:
    """Calculate the coefficient of determination  of two lists of one-dimensional data.

    Coefficient of determination: Explainable variation / total variation. A measure of how much proportional 
    dependent variable variation can be explained by regression lines and independent variables. The closer to 1, 
    the variation explained by this regression line is close to 100%. Algorithm: Square after calculating the linear correlation coefficient.

    Parameters
    ---------
        target_list : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        predict_list : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The coefficient of determination  of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: train_target=",chack_train_target, "; target_list=", chack_target_list,  
        "; predict_list=", chack_predict_list.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_train_target' is'True', it means that 'train_target' does not need to be 
        changed; if it is'False', the input 'train_target' is changed to a one-dimensional list of numerical data. 'target_list' and 'predict_list' have the same judgment and processing 
        method as 'train_target'.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Coefficient of determination: https://zh.wikipedia.org/wiki/%E5%86%B3%E5%AE%9A%E7%B3%BB%E6%95%B0

    """
    try:
        chack_train_target = chack_list_all_num(train_target)
        chack_target_list = chack_list_all_num(target_list)
        chack_predict_list = chack_list_all_num(predict_list)
        if (chack_train_target and chack_target_list and chack_predict_list):
            if(len(target_list) == len(predict_list)):
                train_target_mean = sum(train_target)/len(train_target)
                sst=0.0
                for i in range( len(target_list)):
                    sst=sst+(target_list[i]-train_target_mean)**2
                sst = sst / len(target_list)
                sse = mean_square_error(target_list, predict_list)
                return 1- (sse/sst)
            else:
                raise ValueError("The two lists must be the same length.")
        else:
            raise ValueError("List content must be one numerical data: train_target=",chack_train_target, "; target_list=", chack_target_list,  
                             "; predict_list=", chack_predict_list)
    except TypeError as err:
        raise TypeError(err)

def mean_square_error(target_list, predict_list) -> float:
    """Regression indicators for numerical prediction. Calculate the mean square error (MSE) of two lists of one-dimensional data.

    Mean square error: Measures the mean of the square of the difference between the predicted value and the 
    actual observed value. Algorithm: After subtracting the two lists, the resulting list is multiplied by itself, and finally 
    all the values in the multiplied list are summed, and finally divide by the total number of list elements.

    Parameters
    ---------
        target_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        predict_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The mean square error  of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Mean square error: https://zh.wikipedia.org/wiki/%E5%86%B3%E5%AE%9A%E7%B3%BB%E6%95%B0

    """
    try:
        err = sub(target_list, predict_list)
        mse = sum(mult(err, err))
        return mse/len(target_list)
    except TypeError as err:
        raise TypeError(err)

def mean_absolute_error(target_list, predict_list) -> float:
    """Regression indicators for numerical prediction. Calculate the mean absolute error (MAE) of two lists of one-dimensional data.

     Mean absolute error  can measure the error between actual observations and predicted values. The calculation method is to 
     subtract the two lists, take the absolute value of each number and then add them all, and finally divide by the total number of list elements.

    Parameters
    ---------
        target_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        predict_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The root mean square error  of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    MAE wiki :https://en.wikipedia.org/wiki/Mean_absolute_error

    """
    try:
        mae = sum(np.abs(sub(target_list, predict_list)))
        return mae/len(target_list)
    except TypeError as err:
        raise TypeError(err)

def root_mean_squard_error(target_list, predict_list) -> float:
    """Regression indicators for numerical prediction. Calculate the root mean square error (RMSE) of two lists of one-dimensional data.

    Root mean square error  can measure the error between actual observations and predicted values. 
    The calculation method is the square root of MSE.

    Parameters
    ---------
        target_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        predict_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The root mean square error  of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Root mean square  error(RMSE) wiki: https://en.wikipedia.org/wiki/Root-mean-square_deviation

    """
    try:
        mse = mean_square_error(target_list, predict_list)
        return mse**0.5
    except TypeError as err:
        raise TypeError(err)

def normalized_mean_squard_error(target_list, predict_list) -> float:
    """Regression indicators for numerical prediction. Calculate the normalized mean square error (NRMSE) of two lists of one-dimensional data.

    NRMSE is a statistical value, and its method is to normalize RMSE. The normalization method used by this function is that 
    the error between the maximum and minimum is used as the denominator, and RMSE is used as the numerator.

    The value is between 0 and 1. The closer the NRMSE is to 0, the smaller the error between the two and 
    the closer the model predicted value to the target value.

    Parameters
    ---------
        target_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        predict_list: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The normalized mean square error  of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Normalized mean square error (NRMSE) wiki: https://zh.wikipedia.org/wiki/%E6%AD%A3%E8%A6%8F%E5%8C%96%E6%96%B9%E5%9D%87%E6%A0%B9%E5%B7%AE
    Normalized mean square error (NRMSE) function: https://www.marinedatascience.co/blog/2019/01/07/normalizing-the-rmse/

    """
    try:
        rmse = root_mean_squard_error(target_list, predict_list)
        return rmse/(max(target_list) - min(target_list))
    except TypeError as err:
        raise TypeError(err)

def ACF(listA, listB) -> float:
    """Calculate the autocorrelation coefficients of two lists of one-dimensional data.

    Autocorrelation coefficients: The autocorrelation as a function of the lag. 
    Algorithm: Calculate the autocorrelation coefficients according to the formula.

    Parameters
    ---------
        listA : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        listB : list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    float. The autocorrelation coefficients of two lists of one-dimensional data.

    Error
    ---------
        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

        ValueError: List content must be one-dimensional numerical data: listA=",chack_A, "; listB=", chack_B.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_A' is'True', it means that 'listA' does not need to be 
        changed; if it is'False', the input 'listA' is changed to a one-dimensional list of numerical data.ListB has the same judgment and processing 
        method as listA.

        TypeError: object of type 'NoneType' has no len()
        Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

    References
    ---------
    Autocorrelation coefficients wiki: https://zh.wikipedia.org/wiki/%E8%87%AA%E7%9B%B8%E5%85%B3%E5%87%BD%E6%95%B0
    Autocorrelation Function: https://www.real-statistics.com/time-series-analysis/stochastic-processes/autocorrelation-function/

    """
    try:
        cov_A_B = covariance(listA, listB)
        sd_A = standard_deviation(listA)
        sd_B = standard_deviation(listB)
        P = cov_A_B / (sd_A * sd_B)
        return P
    except TypeError as err:
        raise TypeError(err)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright 2020 by leodflag

"""This module contains the access and processing methods of time series data.

Ways to obtain time series data: read files and download from websites.
The data processing method is archiving, the two formats are interchanged, 
the data is divided, and the time series model data set is established.

Functions
---------
read_file(path: str, col_name: str='close') -> pandas.Series
	Read file to get time series data.

get_data_yahoo(stock_id: str, start_period: str, end_period: str, file_format: str='csv', frequency: str = 'day')
	Crawl and archive the stock price data of finance.yahoo.com.

save_flie(data, path: str, stock_id: str = 'stock', file_format: str = 'csv')
	Save data as file.

create_sequence_list(data) -> list
	Create a sequence starting from 1.

list_to_dataframe(data: list)
	Convert the 'list' type of the one-dimensional list of values to the 'DataFrame' type.

dataframe_to_list(data: pandas.Series)
	Convert the 'pandas.Series' type of the one-dimensional list of values to the'list' type.

split_data(data, ratio: float)
	Split the data into two data sets according to the input ratio.

create_ar_data(data: pandas.Series, lags: int = 1) -> pandas.DataFrame
	Create an autoregressive data set.

n_order_difference_data(data: pandas.Series, periods: int = 1, log: bool = False) -> pandas.Series:
    Sequence of numbers after difference operation.

Notes
-----------
If the time series data type is a list, you can use the function "list_to_dataframe" of the module "TimeSeriesAnalysis.ProcessData" to convert the data type.

"""
import requests
import datetime
import json
import time
import io
import os
import pandas
import numpy as np
import matplotlib.pyplot as plt
from dateutil.parser import parse
import Math.Statistics as Math

def read_file(path: str, col_name: str='close') -> pandas.Series:
    """Read file to get time series data.

    The file formats that can be read by this function are csv and txt, and the returned data type is pandas.Series.

    Parameters
    ---------

        path : string
        Enter the absolute path or relative path where the file is located. The file formats that can be read are csv and txt.

        col_name : string , default = 'close'.
        The text of the file must contain column names.

    Returns
    ---------
    pandas.Series. Return a list with date as the column index value, the data type is float.

    Error
    ---------
        FileNotFoundError : [Errno 2] There is no such file or directory: 'wrong path'
        Solution : Enter the correct file path and the correct file name.

        ValueError: Invalid file path or buffer object type: <class 'int'>
        Solution : Enter the correct file path and the correct file name.

        KeyError : "This file has not index value", err.
        Solution : Actually check whether the column name contained in the file content exists, and enter the correct column name.

    """
    try:
        data = pandas.read_csv(path, index_col='date')
        Data = data[col_name]
        return Data
    except (FileNotFoundError, ValueError) as err:
        print(err)
    except KeyError as err:
        print("This file has not index value", err)

def get_data_yahoo(stock_id: str, start_period: str, end_period: str, file_format: str='csv', frequency: str = 'day'):
    """Crawl and archive the stock price data of finance.yahoo.com.

    This function can crawl the stock market price data at a specified time and frequency and save it as a csv or txt file, and the returned 
    data type is pandas.Series.

    Parameters
    ---------
        stock_id: string
        Enter the stock ID. For example: Taiwan stock" 2206.TW"; US stock "SWKS".

        start_period: string
        The start time of the time period for crawling stock prices.  
        For example: "2019 7 12", "2019, 7, 12", "2019 7", "2019, 7", "2019",  "7 21 2019", "7, 21, 2019", " 7 2019","7, 2019".

        end_period: string
        The start time of the time period for crawling stock prices.  
        For example: "2019 8 12", "2019, 8, 12", "2019 8", "2019, 8", "2019",  "8 21 2019", "8, 21, 2019", " 8 2019","8, 2019".

        file_format: string, default = 'csv'.
        Archive the crawled stock market price data, file format is 'csv' or 'txt'.  For example: 'csv' ,'CSV' , 'txt', 'TXT'.

        frequency: string, default = 'day'.
        The frequency of stock market price data is daily, weekly, and monthly. For example: 'daily', 'day', 'weekly', 'week', 'monthly', 'month'.

    Returns
    ---------
    No return.

    Error
    ---------
        ValueError: 'start_period' must be earlier than 'end_period'.
        Solution: Check whether the parameter matches the 'ValueError' description. For example: start_period = '2020, 2, 5',  end_period = '2020, 4, 5'

        TypeError: Not found stock id.
        Solution: Enter the correct stock id.

        ConnectionError: Internet connection error.
        Solution: Confirm that the network is connected.

    """
    try:
        period1 = parse(start_period)
        period2 = parse(end_period)
        if(period1 > period2 or period1 == period2):
            raise ValueError("'start_period' must be earlier than 'end_period'")

        freq = ''
        if(frequency == 'monthly' or frequency == 'month'):
            freq = 'mo'
        elif (frequency == 'weekly' or frequency == 'week'):
            freq = 'wk'
        elif (frequency == 'daily' or frequency == 'day'):
            freq = 'd'
        else:
            freq = 'd'

        url = ("https://query1.finance.yahoo.com/v7/finance/chart/" + stock_id + "?period1=" + str(int(period1.timestamp())) +
               "&period2=" + str(int(period2.timestamp())) + "&interval=1" + freq + "&events=history")
        req = requests.get(url)
        stock = json.loads(req.text)
        if(stock['chart']['result'] == None):
            raise TypeError("Not found stock id")
        else:
            print("Crawl successfully")
            stock_data = pandas.DataFrame(stock['chart']['result'][0]['indicators']['quote'][0], index=pandas.to_datetime(
                np.array(stock['chart']['result'][0]['timestamp'])*1000*1000*1000))
            if not os.path.isdir(stock_id + '/'):
                os.mkdir(stock_id)
            if (file_format == 'csv' or file_format == 'CSV'):
                stock_data.index = stock_data.index.set_names(['date'])
                stock_data.to_csv(stock_id + '/'+stock_id +'_' + frequency + '.csv')
                print('Saved successfully (get data yahoo). File path = '+ stock_id + '/'+stock_id + '_' + frequency + '.csv')
            elif (file_format == 'txt' or file_format == 'TXT'):
                stock_data.index = stock_data.index.set_names(['date'])
                stock_data.to_csv(stock_id + '/'+stock_id + '_' + frequency + '.txt')
                print('Saved successfully (get data yahoo). File path = '+ stock_id + '/'+stock_id + '_' + frequency + '.txt')
    except requests.exceptions.ConnectionError as err:
        raise requests.exceptions.ConnectionError(err)

def save_flie(data, path: str, stock_id: str = 'stock', file_format: str = 'csv'):
    """Save data as file.

    This function can save the input pandas.Series data type data into a csv or txt file.

    Parameters
    ---------
        data: pandas.Series and pandas.DataFrame.
        pandas.Series: One-dimensional ndarray with axis labels (including time series). 
        pandas.DataFrame: Two-dimensional, size-mutable, potentially heterogeneous tabular data.

        path: str.
        Save files to absolute path or relative path.If you enter a null string '' or "",it will be saved in the current folder

        stock_id: str, default = 'stock'.
        File name when saving the file.

        file_format: str, default = 'csv'.
        Archive the crawled stock market price data, file format is 'csv' or 'txt'.  For example: 'csv' ,'CSV' , 'txt', 'TXT'.

    Returns
    ---------
    No return value.

    Error
    ---------
        PermissionError: Reject operations that do not comply with permissions.
        Solution: Path or file name cannot be '/'.

        AttributeError: 'list' object has no attribute 'to_csv'
        Solution: The data format must be pandas.Series and pandas.DataFrame.

        FileNotFoundError: No such file or file path.
        Solution: Enter the correct file path.

    """
    try:
        if(file_format == 'csv' or file_format == 'CSV'):
            data.to_csv(path + '/' + stock_id + '.csv')
            print('Saved successfully  (csv),  file path =  '+path + '/' + stock_id + '.csv')
        if (file_format == 'txt' or file_format == 'TXT'):
            data.to_csv(path + '/'+stock_id + '.txt')
            print('Saved successfully (txt),  file path =  '+path + '/' + stock_id + '.txt')
    except (PermissionError, AttributeError, FileNotFoundError) as err:
        print(err)

def create_sequence_list(data) -> list:
    """Create a sequence starting from 1.

    This function uses the length of the input one-dimensional numerical list to create a sequence starting from 1.

    Parameters
    ---------
        data: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

    Returns
    ---------
    list. Return a sequence list with the same length as 'data'.

    Error
    ---------
        ValueError: List content must be one-dimensional numerical data: chack_data=",chack_data.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_data' is'True', it means that 'data' does not need to 
        be changed; if it is'False', the input 'data' is changed to a one-dimensional list of numerical data.

        ValueError: 'data' must be one-dimensional numerical list.
        Solution: Check that the input 'data' is a one-dimensional list, for example: [1,5,8,6,3].

    """
    try:
        chack_data = Math.chack_list_all_num(data)
        if(chack_data):
            sequence = []
            data = np.array(data)
            if(data.ndim == 1 and len(data) != 0):
                for i in range(1, (len(data)+1)): sequence.append(i)
                return sequence
        else:
            raise ValueError("List content must be one numerical data: chack_data=", chack_data)
    except :
        raise ValueError("'data' must be one-dimensional numerical list.")

def list_to_dataframe(data: list , index) -> pandas.DataFrame:
    """Convert the 'list' type of the one-dimensional list of values to the 'DataFrame' type.

    Parameters
    ---------
        data: list.
        One-dimensional numerical list. For example: [1,5,8,6,3].

        index: pandas.Index, list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional list.

    Returns
    ---------
    pandas.DataFrame. The one-dimensional list of values to the 'DataFrame' type.

    Error
    ---------
        ValueError: 'data' must be one-dimensional numerical list.
        Solution: 'data' contains a string or is two-dimensional list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        ValueError: The two lists must be the same length.
        Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

    """
    try:
        if (len(data) == len(index)):
            df_data = pandas.DataFrame(data, index = index)
            return df_data
        else:
            raise ValueError("The two lists must be the same length.")
    except :
        raise ValueError("'data' must be one-dimensional numerical list.")

def dataframe_to_list(data: pandas.Series) -> list:
    """Convert the 'pandas.Series' type of the one-dimensional list of values to the'list' type.

    Parameters
    ---------
        data: pandas.Series.
        'data' is a one-dimensional numerical data of type'pandas.Series' taken from'pandas.DataFrame'.

    Returns
    ---------
    list. One-dimensional numerical list.

    Error
    ---------
        ValueError: List content must be one numerical data: chack_data=", chack_data.
        Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
        non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If 'chack_data' is 'True', it means that 'data' does not need to 
        be changed; if it is'False', the input 'data' is changed to a one-dimensional list of numerical data.

        ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
        Solution: Input 'data' must be a one-dimensional numerical data of type'pandas.Series' taken from'pandas.DataFrame'.

    """
    try:
        type_data = str(type(data))
        chack_data = Math.chack_list_all_num(data)
        if(type_data == "<class 'pandas.core.series.Series'>"):
            if(chack_data):
                return np.array(data).tolist()
            else:
                raise ValueError("List content must be one numerical data: chack_data=", chack_data)
        else:
            raise ValueError
    except:
        raise ValueError("The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.")

def split_data(data, ratio: float):
    """Split the data into two data sets according to the input ratio.

    This function splits one-dimensional numerical list into two data sets according to the split ratio. 
    The first data set is the data length of the split ratio, and the second data set is the data length of the split ratio.

    Parameters
    ---------
        data: list ,ndarray, pandas.Series and pandas.DataFrame.
        One-dimensional numerical list.

        ratio: float.
        Limitation factor: 0 < ratio <1. The split ratio is a floating point number used to determine the respective lengths of the two data sets after splitting.

    Returns
    ---------
    Two one-dimensional data sets. The first data set is the data length of the split ratio, and the second data set is the data length of the split ratio.

    Error
    ---------
        TypeError: 'data' of type 'NoneType'.
        Solution: 'data' must be one-dimensional numerical list.

        ValueError: 'ratio' must be between 0 and 1, excluding 0 and 1.
        Solution: Enter a floating point number greater than 0 and less than 1.

        ValueError: There are less than two numerical data in the list.
        Solution: There are at least two numerical data in the list.

    """
    try:
        if(ratio <= 0 or ratio >= 1 ):
            raise ValueError("'ratio' must be between 0 and 1, excluding 0 and 1.")
        else:
            if(len(data) > 2):
                num = int(len(data)*ratio)
                train_data = data[:num]
                test_data = data[-(len(data) - num):]
                return train_data, test_data
            else:
                raise ValueError("There are at least two numerical data in the list.")
    except TypeError as err:
        raise TypeError(err)

def create_ar_data(data: pandas.Series, lags: int = 1) -> pandas.DataFrame:
    """Create an autoregressive data set.

    The autoregressive data set created to facilitate the calculation of autoregressive, its type is'pandas.DataFrame'.
    For example AR(1) data set:
        date　　　　　　　　t-1　　t
        2020-09-29 23:00:00  29.10  28.89
        2020-09-30 23:00:00  28.89  28.91
        2020-10-01 23:00:00  28.91  28.87
        2020-10-04 23:00:00  28.87  28.92
        2020-10-06 13:56:21  28.92  28.79
        -------------------------------------------
        't': Current time series data
        't-1': Previous time series data

    Parameters
    ---------
        data: pandas.Series.
        'data' is a one-dimensional numerical data of type 'pandas.Series' taken from 'pandas.DataFrame'.

        lags: int , default =  1.
        'lags' represents the number of lagging periods. The default value is 1, which means the data is 1 period behind.

    Returns
    ---------
    pandas.DataFrame. The autoregressive data set created to facilitate the calculation of autoregressive.

    Error
    ---------
        ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
        Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

        ValueError: Check that the input list is a one-dimensional numerical data.
        Solution: 'data' is one-dimensional data and the type is'pandas.Series', but the content is not all numeric.

        ValueError: 'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.
        Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

        TypeError: object of type 'NoneType' has no len().
        Solution:  Please do not enter 'None'.Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

        KeyError: The 'data' type is'pandas.DataFrame', please enter the correct key value.
        Solution: Use the correct key value to obtain data.

    """
    try:
        type_data = str(type(data))
        chack_data = Math.chack_list_all_num(data)
        if(type_data != "<class 'pandas.core.series.Series'>"):
            raise ValueError("The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.")
        if (chack_data == False):
            raise ValueError("Check that the input list is a one-dimensional numerical data.")
        if(lags >= 1 and type(lags) == int and lags < len(data)):
            lags_label = []
            for i in range(1, lags+1):
                data_i = data.shift(i)
                if (i == 1):
                    data1 = pandas.concat([data_i, data], axis=1)
                else:
                    data1 = pandas.concat([data_i, data1], axis=1)
            for i in range(lags, 0, -1):
                lags_label.append(('t-'+str(i)))
            lags_label.append('t')
            data1.columns = lags_label
            data1 = data1[lags:]
            return data1
        else:
            raise ValueError("'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.")
    except ValueError as err:
        raise ValueError(err)
    except KeyError:
        raise KeyError("The'data' type is'pandas.DataFrame', please enter the correct key value.")

def n_order_difference_data(data: pandas.Series, periods: int = 1, log: bool = False) -> pandas.Series:
    """Sequence of numbers after difference operation.

    The difference between the current period and the lag period data is called the first-order difference. 
    The difference the data a second time is called the second-order difference.

    Parameters
    ---------
        data: pandas.Series.
        'data' is a one-dimensional numerical data of type 'pandas.Series' taken from 'pandas.DataFrame'.

        periods: int , default =  1.
        'periods' represents the number of lagging periods. The default value is 1, which means the data is 1 period behind.

        log: bool, default  = False
        Does the data need to be multiplied by the natural logarithm.

    Returns
    ---------
    pandas.Series. Return the difference data.

    Error
    ---------
        ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
        Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

        ValueError: Check that the input list is a one-dimensional numerical data.
        Solution: 'data' is one-dimensional data and the type is'pandas.Series', but the content is not all numeric.

        ValueError: 'periods' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.
        Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

        TypeError: object of type 'NoneType' has no len().
        Solution:  Please do not enter 'None'.Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

        KeyError: The 'data' type is'pandas.DataFrame', please enter the correct key value.
        Solution: Use the correct key value to obtain data.

    References
    ---------
    Stationarity and differencing: https://otexts.com/fpp2/stationarity.html
    Time series analysis ppt: http://homepage.ntu.edu.tw/~sschen/Book/Slides/Ch2Basic.pdf

    """
    try:
        if(periods >= 1 and type(periods) == int and periods < len(data)):
            diff_data = []
            ar_data = create_ar_data(data, periods)
            if(log == True):
                data_t_log = Math.log(ar_data['t'])
                data_per_log = Math.log(ar_data['t-'+str(periods)])
                diff_data = Math.sub(data_t_log, data_per_log)
                diff_df = pandas.DataFrame(diff_data, index=ar_data['t'].index)
            elif(log == False):
                diff_data = Math.sub(ar_data['t'], ar_data['t-'+str(periods)])
                diff_df = pandas.DataFrame(diff_data, index = ar_data['t'].index)
            return diff_df[0]
        else:
            raise ValueError("'periods' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.")
    except ValueError as err:
        raise ValueError(err)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright 2020 by leodflag

"""This module contains time series models for forecasting.

Time series prediction models include autoregressive models and moving average models, which can be used for 
model training and model prediction, and model parameters such as prediction results and prediction errors can be 
obtained through the created objects.

Class 1
---------
AutoRegressive
	Autoregressive is a time series model.

Methods
---------
fit(train_data: pandas.Series)
    Fit autoregression model.

predict(test_data: pandas.Series) -> list
    Use an autoregressive model to make predictions.

Class 2
---------
SimpleMovingAverage
	Simple moving average is often used in the financial field..

Methods
---------
fit(train_data: pandas.Series)
    Fit simple moving average.

Notes
-----------
The input data format must be'pandas.Series'.

If the time series data type is a list, you can use the function "list_to_dataframe" of the module "ProcessData" to convert the data type.

It is recommended to use the function of'ProcessData' to obtain and process time series data, which is less prone to errors when entering the model.

"""
import pandas
import numpy
import TimeSeriesAnalysis.ProcessData as Data
import Math.Statistics as Math

class AutoRegressive:

    """Autoregressive is a time series model.

    Autoregressive is a time series model. The model uses past data in the time series as its own explanatory variable, so it will be affected by current errors and past data.

    Parameters
    ---------
        lags: int, default = 1
        Number of lagging periods. Decide to use the data of the previous periods as the explanatory variable. 
        The default value is 1, which means that only the data from the previous period is used as the explanatory variable, which is called the first order autoregressive model (AR(1)).

    Attributes
    ---------
        __data: pandas.Series
        One-dimensional time series data.
        For example:
            date
            2020-06-23 01:00:00    28.89
            2020-06-24 01:00:00    28.91
            2020-06-29 01:00:00    28.87
            2020-06-30 01:00:00    28.92
            2020-07-01 01:00:00    28.79
            Name: close, Length: 170, dtype: float64

        __ar_lags_data: pandas.DataFrame
        Autoregressive data set.
        For example AR(1) data set:
            date　　　　　　　　t-1　　t
            2020-09-29 23:00:00  29.10  28.89
            2020-09-30 23:00:00  28.89  28.91
            2020-10-01 23:00:00  28.91  28.87
            2020-10-04 23:00:00  28.87  28.92
            2020-10-06 13:56:21  28.92  28.79
            -------------------------------------------
            't': Current time series data
            't-1': Previous time series data

        __ar_test_data: pandas.DataFrame
        Autoregressive data set when using 'predict' function.

        __ar_coef_matrix: numpy.ndarray
        Multidimensional linear regression matrix normal equation.

        __ar_train_predict: float
        The error between the current period and the previous period.

        __ar_train_error: list
        Time series forecast results of autoregressive model. The order is [mse, mae, rmse, nrmse].

        __ar_test_data: list
        Time series forecast results of autoregressive model.

        __ar_test_predict: list
        Time series forecast results of autoregressive model.

        __ar_test_error: float
        The mean square error predicted by the autoregressive model. The order is [mse, mae, rmse, nrmse].

        __ar_train_r_2 : float
        The train coefficient of determination .

        __ar_test_r_2 : float
        The test coefficient of determination .

    Methods
    ---------
        lags -> int
            Return the number of lagging periods

        ar_lags_data -> pandas.DataFrame:
        Return autoregressive model data of lagging periods.

        test_data -> pandas.Series
            Return data used for testing.

        coefficients_matrix -> numpy.ndarray
            Return the autoregressive model coefficients matrix.

        train_predict -> pandas.Series
            Return the results predicted by the autoregressive model using the training set.

        test_predict -> pandas.Series
            Return the results predicted by the autoregressive model using the testing set.

        train_error -> list
            Return the regression index used to calculate the training prediction error. The order is [mse, mae, rmse, nrmse].

        test_error -> list
            Return the regression index used to calculate the testing prediction error. The order is [mse, mae, rmse, nrmse].

        ar_train_r_2 -> float
            Return train coefficient of determination .

        ar_test_r_2 -> float
            Return test coefficient of determination .

        fit(train_data: pandas.Series)
            Fit autoregression model.

        predict(test_data: pandas.Series) -> list
            Use an autoregressive model to make predictions.

    Examples
    ---------
        import TimeSeriesAnalysis.ProcessData as Data
        import TimeSeriesAnalysis.Model as model

        col_name = 'close'
        file_path = 'test_data/TWD_unittest.csv'
        times_data_id = 'TWD'
        save_path = 'TWD'

        # TimeSeriesAnalysis.ProcessData
        data = Data.read_file(file_path, col_name)
        train, test = Data.split_data(data, 0.7)

        # TimeSeriesAnalysis.Model
        model = model.AutoRegressive(lags = 2)
        model.fit(train)
        model.predict(test, pure_test_set_predict = True)
        print(model.test_error)  # [0.07378008693836643, 0.23415224307538107, 0.2716249011750698, 1.0278000869996953]

    References
    ---------
    Autoregressive model: https://zh.wikipedia.org/wiki/%E8%87%AA%E8%BF%B4%E6%AD%B8%E6%A8%A1%E5%9E%8B
    Least Squares Estimation of the Parameters: https://slideplayer.com/slide/4238175/14/images/12/12-1+Multiple+Linear+Regression+Models.jpg
    Matrix Approach to Multiple Linear Regression: https://slideplayer.com/slide/4238175/14/images/20/12-1+Multiple+Linear+Regression+Models.jpg

    """

    def __init__(self, lags: int = 1):
        self.__data = []
        self.__lags = lags
        self.__ar_lags_data = []
        self.__ar_coef_matrix = []
        self.__ar_train_predict = []
        self.__ar_train_error = [0,0,0,0] # mse  mae rmse nrmse
        self.__ar_train_r_2 = 0
        self.__ar_test_data = []
        self.__ar_test_predict = []
        self.__ar_test_error = [0, 0, 0, 0]  # mse  mae rmse nrmse
        self.__ar_test_r_2 = 0

    @property
    def lags(self) -> int:
        """Return the number of lagging periods."""
        return self.__lags

    @property
    def ar_lags_data(self) -> pandas.DataFrame:
        """Return autoregressive model data of lagging periods."""
        return self.__ar_lags_data

    @property
    def test_data(self) -> pandas.Series:
        """Return data used for testing."""
        return self.__ar_test_data

    @property
    def coefficients_matrix(self) -> numpy.ndarray:
        """Return the autoregressive model coefficients matrix. """
        return self.__ar_coef_matrix

    @property
    def train_predict(self) -> pandas.Series:
        """Return the results predicted by the autoregressive model using the training set."""
        return self.__ar_train_predict['t']

    @property
    def test_predict(self) -> pandas.Series:
        """Return the results predicted by the autoregressive model using the testing set."""
        return self.__ar_test_predict['t']

    @property
    def train_error(self) -> list:
        """Return the regression index used to calculate the training prediction error. The order is [mse, mae, rmse, nrmse]."""
        return self.__ar_train_error

    @property
    def test_error(self) -> list:
        """Return the regression index used to calculate the testing prediction error. The order is [mse, mae, rmse, nrmse]."""
        return self.__ar_test_error

    @property
    def ar_train_r_2(self) -> float:
        """Return train coefficient of determination ."""
        return self.__ar_train_r_2

    @property
    def ar_test_r_2(self) -> float:
        """Return test coefficient of determination ."""
        return self.__ar_test_r_2

    def __cal_ar_normal_equation(self):
        """Multidimensional linear regression matrix normal equation."""
        leng = self.__lags + 1
        A_matrix = numpy.zeros([leng, leng])
        B_matrix = numpy.zeros([leng, 1])
        for i in range(leng):
            for j in range(leng):
                if (i == 0 and j == 0):
                    A_matrix[i][j] = len(self.__ar_lags_data['t'])
                    B_matrix[i][0] = self.__ar_lags_data['t'].sum()
                elif (i == 0 and i != j):
                    A_matrix[i][j] = self.__ar_lags_data['t-' + str(j)].sum()
                    B_matrix[j][0] = (self.__ar_lags_data['t']*self.__ar_lags_data['t-' + str(j)]).sum()
                elif (j == 0 and j !=i ):
                    A_matrix[i][j] = self.__ar_lags_data['t-' + str(i)].sum()
                else:
                    A_matrix[i][j] = (self.__ar_lags_data['t-'+str(i)] * self.__ar_lags_data['t-'+str(j)]).sum()
        A_inv = numpy.linalg.inv(A_matrix)
        coef_matrix = A_inv.dot(B_matrix)
        return coef_matrix

    def __predict_train(self):
        """Use an autoregressive model to predict the training set data."""
        predict = []
        for day in range(len(self.__ar_lags_data['t'])):
            pre = self.__ar_coef_matrix[0][0]
            for i in range(1, len(self.__ar_coef_matrix)):
                pre += self.__ar_coef_matrix[i][0]*self.__ar_lags_data['t-'+str(i)][day]
            predict.append(pre)
        self.__ar_train_predict = pandas.DataFrame({ 't': predict}, index = self.__ar_lags_data['t'].index)
        self.__ar_train_error[0] = Math.mean_square_error(self.__ar_lags_data['t'], self.__ar_train_predict['t'])
        self.__ar_train_error[1] = Math.mean_absolute_error(self.__ar_lags_data['t'], self.__ar_train_predict['t'])
        self.__ar_train_error[2] = Math.root_mean_squard_error(self.__ar_lags_data['t'], self.__ar_train_predict['t'])
        self.__ar_train_error[3] = Math.normalized_mean_squard_error(self.__ar_lags_data['t'], self.__ar_train_predict['t'])
        self.__ar_train_r_2 = Math.coefficient_of_determination(self.__data, self.__ar_lags_data['t'], self.__ar_train_predict['t'])

    def fit(self, train_data: pandas.Series):
        """Fit autoregression model.

        Parameters
        ---------
            train_data: pandas.Series.
            Data for training the autoregression model.'train_data' is a one-dimensional numerical data of type'pandas.Series' taken from'pandas.DataFrame'.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            ValueError: 'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'. The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            KeyError: The 'train_data' type is'pandas.DataFrame', please enter the correct key value.
            Solution: Use the correct key value to obtain data.

        """
        self.__data = train_data
        self.__ar_lags_data = Data.create_ar_data(self.__data, self.__lags)
        self.__ar_coef_matrix = self.__cal_ar_normal_equation()
        self.__predict_train()

    def predict(self, test_data: pandas.Series, pure_test_set_predict: bool = True) -> pandas.Series:
        """Use an autoregressive model to make predictions.

        Parameters
        ---------
            test_data: pandas.Series.
            Data for testing the autoregression model.'test_data' is a one-dimensional numerical data of type'pandas.Series' taken from'pandas.DataFrame'.

            pure_test_set_predict: bool, default = True.
            The autoregressive model consumes data with a length of 'lags' when predicting.
            'pure_test_set_predict' default is true, which means that only the data of the test set is used for prediction, so the predicted result will be less than the number of'lags' data. 
            If the setting is false, the end data of the training set will be used as the starting point of the test set prediction, so the predicted data length will be the same as the test set data length.
            If 'pure_test_set_predict' is false, the end data of the training set will be used as the starting point of the test set prediction, so the predicted data length will be the same as the test set data length.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'. The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            KeyError: The 'test_data' type is'pandas.DataFrame', please enter the correct key value.
            Solution: Use the correct key value to obtain data.

        """

        if(pure_test_set_predict == True):
            predict = Data.dataframe_to_list(test_data[:self.__lags])
            for day in range(len(test_data[self.__lags:])):
                pre =  self.__ar_coef_matrix[0][0]
                for i in range(self.__lags, 0, -1):
                    pre += self.__ar_coef_matrix[i][0]*predict[day]
                    day += 1
                day -= self.__lags
                predict.append(pre)
            self.__ar_test_data = test_data[self.__lags:]
            self.__ar_test_predict = pandas.DataFrame({ 't': predict[self.__lags:]}, index = test_data[self.__lags:].index)
            self.__ar_test_error[0] = Math.mean_square_error(test_data[self.__lags:], self.__ar_test_predict['t'])
            self.__ar_test_error[1] = Math.mean_absolute_error(test_data[self.__lags:], self.__ar_test_predict['t'])
            self.__ar_test_error[2] = Math.root_mean_squard_error(test_data[self.__lags:], self.__ar_test_predict['t'])
            self.__ar_test_error[3] = Math.normalized_mean_squard_error(test_data[self.__lags:], self.__ar_test_predict['t'])
            self.__ar_test_r_2 = Math.coefficient_of_determination(self.__data, test_data[self.__lags:], self.__ar_test_predict['t'])
        elif(pure_test_set_predict == False):
            predict = Data.dataframe_to_list(self.__ar_lags_data['t'][-self.__lags:])
            for day in range(len(test_data)):
                pre =  self.__ar_coef_matrix[0][0]
                for i in range(self.__lags, 0, -1):
                    pre += self.__ar_coef_matrix[i][0]*predict[day]
                    day += 1
                day -= self.__lags
                predict.append(pre)
            self.__ar_test_data = test_data
            self.__ar_test_predict = pandas.DataFrame({ 't': predict[self.__lags:]}, index = test_data.index)
            self.__ar_test_error[0] = Math.mean_square_error(test_data, self.__ar_test_predict['t'])
            self.__ar_test_error[1] = Math.mean_absolute_error(test_data, self.__ar_test_predict['t'])
            self.__ar_test_error[2] = Math.root_mean_squard_error(test_data, self.__ar_test_predict['t'])
            self.__ar_test_error[3] = Math.normalized_mean_squard_error(test_data, self.__ar_test_predict['t'])
            self.__ar_test_r_2 = Math.coefficient_of_determination(self.__data, test_data, self.__ar_test_predict['t'])

class SimpleMovingAverage:

    """Simple moving average is often used in the financial field.

    The calculation method is to calculate the average after adding n previous data.

    Parameters
    ---------
        windows: int, default = 5
        Number of lagging periods. Decided to use data from previous periods as explanatory variables.
        The default value is 5, which means that only the data of the previous period is used as an explanatory variable, which is called the simple moving average (SMA(5)).

    Attributes
    ---------
        __data: pandas.Series
        One-dimensional time series data.
        For example:
            date
            2020-06-23 01:00:00    28.89
            2020-06-24 01:00:00    28.91
            2020-06-29 01:00:00    28.87
            2020-06-30 01:00:00    28.92
            2020-07-01 01:00:00    28.79
            Name: close, Length: 170, dtype: float64

        __windows: int, default = 5
        Determine the number of windows containing n lag periods.

        __sma_data: One-dimensional array
        Simple moving average data.
        For example:
            date
            2020-09-29 23:00:00    29.10
            2020-09-30 23:00:00    28.89
            2020-10-01 23:00:00    28.91
            2020-10-04 23:00:00    28.87
            2020-10-06 13:56:21    28.92
            Name: 0, dtype: float64

    Methods
    ---------
        data -> pandas.Series:
            Return the one-dimensional time series data.

        sma_result -> pandas.Series
            Return the result of a one-dimensional number list after simple moving average calculation.

        fit(train_data: pandas.Series)
            Fit simple moving average.

    Examples
    ---------
        import TimeSeriesAnalysis.ProcessData as Data
        import TimeSeriesAnalysis.Model as model

        col_name = 'close'
        file_path = 'Stock.csv'
        times_data_id = 'Stock.TW'

        data = Data.read_file(file_path, col_name)

        model = Model.SimpleMovingAverage(windows= 5)
        model.fit(data)
        print(model.sma_data)

    References
    ---------
    Simple moving average: https://en.wikipedia.org/wiki/Moving_average#Simple_moving_average

    """

    def __init__(self, windows: int = 5):
        self.__data = []
        self.__windows = windows
        self.__sma_result = []

    @property
    def data(self) -> pandas.Series:
        """Return the one-dimensional time series data."""
        return self.__data

    @property
    def sma_result(self) -> pandas.Series:
        """Return the result of a one-dimensional number list after simple moving average calculation."""
        return self.__sma_result[0]

    def __cal_sma_result(self):
        """Calculate simple moving average."""
        if(self.__windows < 2 or self.__windows > len(self.__data)):
            raise ValueError("'windows' > 2 and 'windows' <= data.lenght")
        sma_d = []
        for i in range(0,len(self.__data)-self.__windows+1):
            Sum = 0
            for w in range(self.__windows):
                Sum += self.__data[i+w]
            sma_d.append(Sum/self.__windows)
        self.__sma_result = pandas.DataFrame(sma_d, index = self.__data[self.__windows-1:].index)

    def fit(self, data: pandas.Series):
        """Fit simple moving average.

        Parameters
        ---------
            data: pandas.Series.
            Calculate simple moving average.'data' is a one-dimensional numerical data of type'pandas.Series' taken from'pandas.DataFrame'.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            ValueError: 'windows' > 2 and 'windows' <= data.lenght
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'. The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            KeyError: The 'train_data' type is'pandas.DataFrame', please enter the correct key value.
            Solution: Use the correct key value to obtain data.

        """
        self.__data = data
        self.__cal_sma_result()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright 2020 by leodflag

"""Time series data related drawing function.

The drawing function of this module enables time series data to be presented in a 
chart and automatically saved under the folder name entered when creating the category.

Class
---------
chart
    When creating objects of this category, you can define content settings such as chart title, font, and chart size.

Methods
---------
historocal_trend_line_chart(self, data, file_path: str, xlabel: str = 'date', ylabel: str = 'price')
    Draw historical trend line chart with time series data. After the drawing is completed, it will be saved to the specified path 'file_path'.

line_chart(self, data, predict_data, chart_title: str, file_path: str, xlabel: str = 'date', ylabel: str = 'price')
    Throw in two lists of one-dimensional numbers and draw a line chart, it will be saved to the specified path 'file_path'.

lag_plot(data: pandas.Series, file_path: str)
    Scatter plot of lagging periods.

ACF_chart(data: pandas.Series, file_path: str, lags: int = 1)
    Autocorrelation coefficient chart.

forecast_result_chart_model(self, data, model: Model, file_path: str, model_name: str = "model", xlabel='date', ylabel='price')
    The overall prediction results of a model using the training set and the test set.

forecast_result_chart_predict(self, model: Model, file_path: str, model_name: str = "model", xlabel='date', ylabel='price')
    The prediction result of a model using the test set.

statistics_infographic(self, data: pandas.Series, file_path: str, lags: int = 1, xlabel: str = 'date', ylabel: str = 'price')
    Statistical graphs of time series data.

forecast_result_group_chart(self,train, test, model_1: Model, model_2: Model, file_path: str, model_1_name: str = "model1", model_2_name: str = "model2", xlabel: str = 'date', ylabel: str = 'price')
    Combine and compare the prediction results of the two models.

Notes
-----------
Pay attention to the input data type of the function.

Use the function "list_to_dataframe" or "dataframe_to_list" of the module "TimeSeriesAnalysis.ProcessData" to convert the data type.

"""
import os
import pandas
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import Math.Statistics as Math
import TimeSeriesAnalysis.ProcessData as Data
import TimeSeriesAnalysis.Model as Model

class chart:

    """Time series chart.

    Parameters
    ---------
        title_times_data: string, default = 'times_data'
        Titles of all time series charts.

    Attributes
    ---------
        __figsize_length: int, default = 9
        The length of the chart.

        __figsize_width: int, default = 6
        The width of the chart.

        __fontsize_title: int, default = 15
        Font size of chart title.

        __fontsize_x_y: int, default = 8
        The font size of the chart x-axis and y-axis scale.

        __fontsize_x_y_label: int, default = 11
        The font size of the chart's x-axis and y-axis names.

        self.__data_lag_R: float, default = 0
        The list contains correlation coefficient.

        self.__model1_result: list, default = [0,0,0, 0]
        The list contains mean_square_error、mean_absolute_error、root_mean_squard_error and normalized_mean_squard_error.

        self.__model2_result: list, default = [0, 0, 0, 0]
        The list contains mean_square_error、mean_absolute_error、root_mean_squard_error and normalized_mean_squard_error.

    Methods
    ---------

        lag_plot_R -> dict
            Return correlation coefficient.

        forecast_result_group_model1_evaluation-> dict
            Return model1 mse、mae、rmse、nrmse.

        forecast_result_group_model2_evaluation -> dict
            Return model2 mse、mae、rmse、nrmse.

        n_order_difference_graph(self, data, file_path: str, periods: str = 'First', log: bool = False, xlabel: str = 'date')
            Draw n-order difference graph with time series data.

        historocal_trend_line_chart(self, data, file_path: str, xlabel: str = 'date', ylabel: str = 'price')
            Draw historical trend line chart with time series data. After the drawing is completed, it will be saved to the specified path 'file_path'.

        line_chart(self, data, predict_data, chart_title: str, file_path: str, xlabel: str = 'date', ylabel: str = 'price')
            Throw in two lists of one-dimensional numbers and draw a line chart, it will be saved to the specified path 'file_path'.

        lag_plot(data: pandas.Series, file_path: str)
            Scatter plot of lagging periods.

        ACF_chart(data: pandas.Series, file_path: str, lags: int = 1)
            Autocorrelation coefficient chart.

        forecast_result_chart_model(self, data, model: Model, file_path: str, model_name: str = "model", xlabel='date', ylabel='price')
            The overall prediction results of a model using the training set and the test set.

        forecast_result_chart_predict(self, model: Model, file_path: str, model_name: str = "model", xlabel='date', ylabel='price')
            The prediction result of a model using the test set.

        statistics_infographic(self, data: pandas.Series, file_path: str, lags: int = 1, xlabel: str = 'date', ylabel: str = 'price')
            Statistical graphs of time series data.

        forecast_result_group_chart(self,train, test, model_1: Model, model_2: Model, file_path: str, model_1_name: str = "model1", model_2_name: str = "model2", xlabel: str = 'date', ylabel: str = 'price')
            Combine and compare the prediction results of the two models.

    Examples
    ---------
        import TimeSeriesAnalysis.ProcessData as Data
        import TimeSeriesAnalysis.Chart as Chart
        import TimeSeriesAnalysis.Model as model

        col_name = 'close'
        file_path = 'Stock.csv'
        times_data_id = 'Stock.TW'
        save_path = 'Stock_file'

        data = Data.read_file(file_path, col_name)

        chart = Chart.chart(times_data_id)
        chart.historocal_trend_line_chart(data, save_path)
        chart.lag_plot(data, save_path)
        chart.ACF_chart(data, save_path, 8)

        train, test = Data.split_data(data, 0.7)
        model = model.AutoRegressive()
        model.fit(train)
        model.predict(test)

        chart.forecast_result_line_graph(model.test_data, model.prediction, save_path, 'AR')

    """

    def __init__(self, title_times_data: str = 'times_data', figsize_length: int = 9, figsize_width: int = 6,
                 fontsize_title: int = 15, fontsize_x_y: int = 8, fontsize_x_y_label: int =11):
        self.__title_times_data = title_times_data
        self.__figsize = (figsize_length, figsize_width)
        self.__fontsize_title = fontsize_title
        self.__fontsize_x_y = fontsize_x_y
        self.__fontsize_x_y_label = fontsize_x_y_label
        self.__data_lag_R = 0
        self.__model1_result = [0,0,0, 0, 0]
        self.__model2_result = [0, 0, 0, 0, 0]

    @property
    def lag_plot_R(self) -> dict:
        """Return correlation coefficient."""
        dic_lag = {'R': self.__data_lag_R}
        return dic_lag

    @property
    def forecast_result_group_model1_evaluation(self) -> dict:
        """Return model1 mse、mae、rmse、nrmse."""
        dic_forecast = {'mse': self.__model1_result[0], 'mae': self.__model1_result[1],  'rmse': self.__model1_result[2], 'nrmse': self.__model1_result[3]}
        return dic_forecast

    @property
    def forecast_result_group_model2_evaluation(self) -> dict:
        """Return model2 mse、mae、rmse、nrmse."""
        dic_forecast = {'mse': self.__model2_result[0], 'mae': self.__model2_result[1], 'rmse': self.__model2_result[2],  'nrmse': self.__model2_result[3]}
        return dic_forecast

    def n_order_difference_graph(self, data, file_path: str, periods: str = 'First', log: bool = False, xlabel: str = 'date'):
        """Draw n-order difference graph  with time series data.

        For example: First-order difference graph or Second-order difference graph. After the drawing is completed, it will be saved to the specified path 'file_path'.

        Parameters
        ---------
            data: list ,ndarray, pandas.Series and pandas.DataFrame.
            One-dimensional numerical list.

            file_path: string
            Path to save image file.

            periods: str, default  = 'First'
            Differential period.

            log: bool, default  = False
            The data multiplied by the natural logarithm. True or False.

            xlabel: str, default  = 'date'
            The name of the x-axis.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: ('List content must be one numerical data: data=', False)
            Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
            non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_data' is'True', it means that 'data' does not need to be 
            changed; if it is'False', the input 'data' is changed to a one-dimensional list of numerical data.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The entered'listA' is an empty list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

            KeyError: "The'data' type is'pandas.DataFrame', please enter the correct key value."
            Solution: Use the correct key value to obtain data.

        """
        try:
            chack_data = Math.chack_list_all_num(data)
            str_log = ''
            if (log == True):
                str_log = 'log'
            if(chack_data):
                plt.figure(figsize = self.__figsize)
                plt.title(periods + " order difference " + str_log + " : "+ self.__title_times_data, fontsize=self.__fontsize_title)
                plt.xlabel(xlabel, fontsize = self.__fontsize_x_y_label)
                plt.xticks(np.linspace(0, len(data), 12), rotation=15)
                plt.plot(data)
                plt.tight_layout()
                if not os.path.isdir(file_path + '/'):
                    os.mkdir(file_path)
                plt.savefig(file_path + "/" + periods + " order difference" + str_log + " : "+ self.__title_times_data + ".png")
                plt.show()
                print('Saved successfully (n order difference). File path = ' +
                      file_path + "/" + periods + "_order_difference_" + str_log +'_'+ self.__title_times_data + ".png")
            else:
                raise ValueError("List content must be one numerical data: data=",chack_data)
        except TypeError as err:
            raise TypeError(err)

    def historocal_trend_line_chart(self, data, file_path: str, xlabel: str = 'date', ylabel: str = 'price'):
        """Draw historical trend line chart with time series data. After the drawing is completed, 
        it will be saved to the specified path 'file_path'.

        Parameters
        ---------
            data: list ,ndarray, pandas.Series and pandas.DataFrame.
            One-dimensional numerical list.

            file_path: string
            Path to save image file.

            xlabel: str, default  = 'date'
            The name of the x-axis.

            ylabel: str, default  = 'price'
            The name of the y-axis

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: ('List content must be one numerical data: data=', False)
            Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
            non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_data' is'True', it means that 'data' does not need to be 
            changed; if it is'False', the input 'data' is changed to a one-dimensional list of numerical data.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The entered'listA' is an empty list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

            KeyError: "The'data' type is'pandas.DataFrame', please enter the correct key value."
            Solution: Use the correct key value to obtain data.

        """
        try:
            chack_data = Math.chack_list_all_num(data)
            if(chack_data):
                plt.figure(figsize = self.__figsize)
                plt.title(" Historocal trend : " + self.__title_times_data, fontsize=self.__fontsize_title)
                plt.xlabel(xlabel, fontsize = self.__fontsize_x_y_label)
                plt.ylabel(ylabel, fontsize = self.__fontsize_x_y_label)
                plt.tick_params(axis = 'both', labelsize = self.__fontsize_x_y)
                plt.xticks(np.linspace(0, len(data), 12), rotation=15)
                plt.plot(data)
                plt.tight_layout()
                if not os.path.isdir(file_path + '/'):
                    os.mkdir(file_path)
                plt.savefig(file_path + "/Historocal_Trend_" + self.__title_times_data + ".png")
                plt.show()
                print('Saved successfully (historocal trend). File path = ' +
                      file_path + "/Historocal_Trend_" + self.__title_times_data + ".png")
            else:
                raise ValueError("List content must be one numerical data: data=",chack_data)
        except TypeError as err:
            raise TypeError(err)

    def line_chart(self, data, predict_data, chart_title: str, file_path: str, xlabel: str = 'date', ylabel: str = 'price'):
        """Throw in two lists of one-dimensional numbers and draw a line chart, it will be saved to the specified path 'file_path'.

        Parameters
        ---------
            data: list ,ndarray, pandas.Series and pandas.DataFrame.
            One-dimensional numerical list.

            predict_data: list ,ndarray, pandas.Series and pandas.DataFrame.
            One-dimensional numerical list.

            chart_title: string
            Chart title.

            file_path: string
            Path to save image file.

            xlabel: str, default  = 'date'
            The name of the x-axis.

            ylabel: str, default  = 'price'
            The name of the y-axis

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: ('List content must be one numerical data: data=', False)
            Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
            non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_data' is'True', it means that 'data' does not need to be 
            changed; if it is'False', the input 'data' is changed to a one-dimensional list of numerical data.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The entered'listA' is an empty list. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

            KeyError: "The'data' type is'pandas.DataFrame', please enter the correct key value."
            Solution: Use the correct key value to obtain data.

        """
        try:
            chack_data = Math.chack_list_all_num(data)
            chack_pre = Math.chack_list_all_num(predict_data)
            if(chack_data and chack_pre):
                plt.figure(figsize = self.__figsize)
                plt.title(chart_title + ' :' + self.__title_times_data, fontsize=self.__fontsize_title)
                plt.xlabel(xlabel, fontsize = self.__fontsize_x_y_label)
                plt.ylabel(ylabel, fontsize = self.__fontsize_x_y_label)
                plt.tick_params(axis = 'both', labelsize = self.__fontsize_x_y)
                plt.xticks(np.linspace(0, len(data), 12), rotation=15)
                plt.plot(data, color='lightgray')
                plt.plot(predict_data, color = 'green')
                plt.tight_layout()
                if not os.path.isdir(file_path + '/'):
                    os.mkdir(file_path)
                plt.savefig(file_path + "/" + chart_title + "_" + self.__title_times_data + "_line_chart.png")
                plt.show()
                print('Saved successfully (historocal trend analysis). File path = ' +
                      file_path + "/" + chart_title + "_" + self.__title_times_data + "_line_chart.png")
            else:
                raise ValueError("List content must be one numerical data: data=",chack_data, " predict_data=", chack_pre)
        except TypeError as err:
            raise TypeError(err)

    def lag_plot(self, data: pandas.Series, file_path: str):
        """Scatter plot of lagging periods.

        The current time series data and the previous time series data are drawn into a scatter diagram to determine the linear relationship.
        Current data is the x-axis, previous data is the y-axis.After the drawing is completed, it will be saved to the specified path 'file_path'.

        Parameters
        ---------
            data: pandas.Series.
            'data' is a one-dimensional numerical time series data of type'pandas.Series' taken from'pandas.DataFrame'.

            file_path: string
            Path to save image file.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The entered'data' is an empty list. Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            ValueError: 'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.
            Solution: Check whether the parameter matches the 'ValueError' description.

            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'.Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

        """
        try:
            plt.figure(figsize=self.__figsize)
            lags_data = Data.create_ar_data(data)
            x_site = min(lags_data['t'])
            y_site = min(lags_data['t-1'])
            self.__data_lag_R = round(Math.correlation_coefficient(lags_data['t'], lags_data['t-1']), 4)
            plt.title("Lag Plot : "+self.__title_times_data, fontsize=self.__fontsize_title)
            plt.xlabel("y(t)", fontsize=self.__fontsize_x_y_label)
            plt.ylabel("y(t-1)", fontsize=self.__fontsize_x_y_label)
            plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
            plt.text(x_site, y_site, r'$\mathrm{r} = $' + str(self.__data_lag_R),
                     bbox=dict( boxstyle="square", ec=(1., 0.5, 0.5), fc=(1., 0.9, 0.9), ), fontsize=self.__fontsize_x_y_label)
            plt.scatter(lags_data['t'], lags_data['t-1'], s = 10)
            plt.tight_layout()
            if not os.path.isdir(file_path + '/'):
                os.mkdir(file_path)
            plt.savefig(file_path + "/Lag_Plot_" + self.__title_times_data + ".png")
            plt.show()
            print('Saved successfully (lag plot). File path = ' + file_path + "/Lag_Plot_" + self.__title_times_data + ".png")
        except AttributeError as err:
            print(err)

    def ACF_chart(self, data: pandas.Series, file_path: str, lags: int = 1):
        """Autocorrelation coefficient chart.

        Compare a time series data with itself. Bar graph. The number of deferral periods is on the x-axis, and ACF is on the y-axis
        The value range of the obtained autocorrelation value R is [-1,1], 1 is the maximum positive correlation 
        value,-1 is the maximum negative correlation value, and 0 is irrelevant. After the drawing is completed, it will be saved to the specified path 'file_path'.

        Parameters
        ---------
            data: pandas.Series
            'data' is a one-dimensional numerical time series data of type'pandas.Series' taken from'pandas.DataFrame'.

            file_path: string
            Path to save image file.

            lags: int, default  = 1
            'lags' represents the number of lagging periods. The default value is 1, which means the data is 1 period behind.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The entered'data' is an empty list. Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            ValueError: 'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.
            Solution: Check whether the parameter matches the 'ValueError' description.

            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'.Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

        """
        try:
            if(lags >= 1 and type(lags) == int and lags < len(data)):
                ar_data = Data.create_ar_data(data, lags)
                acf = []
                for i in range(1,lags+1):
                    acf_value = 0
                    acf_value = Math.ACF(ar_data['t'], ar_data['t-'+str(i)])
                    acf.append(acf_value)
                index_list = Data.create_sequence_list(acf)
                df_acf = Data.list_to_dataframe(acf, index_list)
                plt.figure(figsize=self.__figsize)
                plt.title("ACF : "+self.__title_times_data, fontsize=self.__fontsize_title)
                plt.xlabel("lags", fontsize=self.__fontsize_x_y_label)
                plt.ylabel("ACF", fontsize=self.__fontsize_x_y_label)
                plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
                plt.bar(df_acf.index, df_acf[0])
                plt.tight_layout()
                if not os.path.isdir(file_path + '/'):
                    os.mkdir(file_path)
                plt.savefig(file_path + "/ACF_" + self.__title_times_data + ".png")
                plt.show()
                print('Saved successfully (ACF). File path = ' + file_path + "/ACF_" + self.__title_times_data + ".png")
            else:
                raise ValueError("'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.")
        except AttributeError as err:
            raise AttributeError(err)

    def forecast_result_chart_model(self, data, model: Model, file_path: str, model_name: str = "model", xlabel='date', ylabel='price'):
        """The overall prediction results of a model using the training set and the test set.

        The graph drawn by this function can display historical data, the prediction results of the training set 
        and the test set at the same time, as well as the model measurement indicators, such as MSE, MAE, RMSE, 
        NRMSE, and finally saved in the folder.

        Parameters
        ---------
            data: list ,ndarray, pandas.Series and pandas.DataFrame.
            One-dimensional numerical list.

            model: Model
            The object type is TimeSeriesAnalysis.Model.

            file_path: string
            Path to save image file.

            model_name: str, default  = "model"
            The name of the model.

            xlabel: str, default  = 'date'
            The name of the x-axis.

            ylabel: str, default  = 'price'
            The name of the y-axis

        Returns
        ---------
        No return value.

        Error
        ---------
            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'.Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

        """
        try:
            plt.figure(figsize=self.__figsize)
            plt.title("Forecast Result Model (" +  model_name + "): " + self.__title_times_data, fontsize=self.__fontsize_title)
            plt.plot(data, color='lightgray', label='actul')
            plt.plot(model.train_predict, label="train predict")
            plt.plot(model.test_predict, color="red", label="test predict")
            plt.xlabel(xlabel, fontsize=self.__fontsize_x_y_label)
            plt.ylabel(ylabel, fontsize=self.__fontsize_x_y_label)
            x_site = min(data)
            y_site = min(data)
            plt.text(x_site*0.05, y_site, r'$\mathrm{train} : $' + '\n\n' +
                     r'$\mathrm{MSE} = $' + str(round(model.train_error[0], 4)) + '\n' +
                     r'$\mathrm{MAE} = $' + str(round(model.train_error[1], 4)) + '\n' +
                     r'$\mathrm{RMSE} = $' + str(round(model.train_error[2], 4)) + '\n' +
                     r'$\mathrm{NRMSE} = $' + str(round(model.train_error[3], 4)) 
                     + '\n\n' + r'$\mathrm{test} : $' + '\n\n' + r'$\mathrm{MSE} = $' + str(round(model.test_error[0], 4)) 
                      +'\n' + r'$\mathrm{MAE} = $' + str(round(model.test_error[1], 4)) + '\n' + r'$\mathrm{RMSE} = $' + str(round(model.test_error[2], 4)) 
                      + '\n' + r'$\mathrm{NRMSE} = $' + str(round(model.test_error[3], 4)),
                            bbox=dict( boxstyle="square", ec=(1., 0.5, 0.5), fc=(1., 0.9, 0.9), ), fontsize=self.__fontsize_x_y)
            plt.xticks(np.linspace(0, len(data), 12), rotation=15)
            plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
            plt.legend(loc='upper left')
            plt.tight_layout()
            if not os.path.isdir(file_path + '/'):
                os.mkdir(file_path)
            plt.savefig(file_path + "/Forecast_Result_Model_"+  model_name +"_" +  self.__title_times_data + ".png")
            plt.show()
            print('Saved successfully (forecast result model '+ model_name +'). File path = ' + file_path +
                    "/Forecast_Result_Model" + model_name + "_" + self.__title_times_data + ".png")
        except TypeError as err:
            raise TypeError(err)

    def forecast_result_chart_predict(self, model: Model, file_path: str, model_name: str = "model", xlabel='date', ylabel='price'):
        """Forecast result graph.

        Input test data and predicted results are drawn as a line graph for comparison.After the drawing is completed, it will be saved to the specified path 'file_path'.

        Parameters
        ---------
            model: Model
            The object type is TimeSeriesAnalysis.Model.

            file_path: string
            Path to save image file.

            model_name: str, default   = "model"
            The name of the time series model.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The two lists must be the same length.
            Solution: Make sure that the number of numerical data in the two lists is the same (the same length).

            ValueError: List content must be one-dimensional numerical data:  test_data=", chack_test, "; predict_data=", chack_pre.
            Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
            non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_test' is'True', it means that 'test_data' does not need 
            to be changed; if it is'False', the input 'test_data' is changed to a one-dimensional list of numerical data. predict_data has 
            the same judgment and processing method as test_data.

            TypeError: object of type 'NoneType' has no len()
            Solution: 'listA' or 'listB' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        """
        try:
            plt.figure(figsize=self.__figsize)
            plt.title("Forecast Result " +  model_name + " : " + self.__title_times_data, fontsize=self.__fontsize_title)
            plt.plot(model.test_data, color='lightgray', label='actul')
            plt.plot(model.test_predict, color="red", label="test predict")
            plt.xlabel(xlabel, fontsize=self.__fontsize_x_y_label)
            plt.ylabel(ylabel, fontsize=self.__fontsize_x_y_label)
            x_site = len(model.test_data)
            y_site = min(model.test_data)
            plt.text(x_site*0.05, y_site, r'$\mathrm{test} : $' + '\n\n' + r'$\mathrm{MSE} = $' + str(round(model.test_error[0], 4)) 
                      +'\n' + r'$\mathrm{MAE} = $' + str(round(model.test_error[1], 4)) + '\n' + r'$\mathrm{RMSE} = $' + str(round(model.test_error[2], 4)) 
                      + '\n' + r'$\mathrm{NRMSE} = $' + str(round(model.test_error[3], 4)),
                            bbox=dict( boxstyle="square", ec=(1., 0.5, 0.5), fc=(1., 0.9, 0.9), ), fontsize=self.__fontsize_x_y)
            plt.xticks(np.linspace(0, len(model.test_data), 12), rotation=15)
            plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
            plt.legend(loc='best')
            plt.tight_layout()
            if not os.path.isdir(file_path + '/'):
                os.mkdir(file_path)
            plt.savefig(file_path + "/Forecast_Result_predict_"+  model_name +"_" +  self.__title_times_data + ".png")
            plt.show()
            print('Saved successfully (forecast result presict '+ model_name +'). File path = ' + file_path +
                    "/Forecast_Result_predict_" + model_name + "_" + self.__title_times_data + ".png")
        except TypeError as err:
            raise TypeError(err)

    def statistics_infographic(self, data: pandas.Series, file_path: str, lags: int = 1, xlabel: str = 'date', ylabel: str = 'price'):
        """Statistical graphs of time series data.

    Three statistical graphs combined into one graph: historocal trend line chart、lag plot、ACF chart.

        Parameters
        ---------
            data: pandas.Series
            'data' is a one-dimensional numerical time series data of type'pandas.Series' taken from'pandas.DataFrame'.

            file_path: string
            Path to save image file.

            lags: int, default  = 1
            'lags' represents the number of lagging periods. The default value is 1, which means the data is 1 period behind.

            xlabel: str, default  = 'date'
            The name of the x-axis for historocal trend line chart.

            ylabel: str, default  = 'price'
            The name of the y-axis for historocal trend line chart.

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: The type of one-dimensional data must be 'pandas.Series', and the content is numeric data.
            Solution: Check whether the parameter matches the 'ValueError' description, or use the default value.

            ValueError: The list must be one-dimensional numerical data, and there is at least one numerical data in it.
            Solution: The entered'data' is an empty list. Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

            ValueError: 'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.
            Solution: Check whether the parameter matches the 'ValueError' description.

            TypeError: object of type 'NoneType' has no len().
            Solution:  Please do not enter 'None'.Check that the type of one-dimensional data must be 'pandas.Series', and the content is numeric data.

        """
        try:
            if(lags >= 1 and type(lags) == int and lags < len(data)):
                chack_data = Math.chack_list_all_num(data)
                if(chack_data):
                    plt.figure(figsize=(15,9))
                    G = gridspec.GridSpec(2, 4)
                    ar_data = Data.create_ar_data(data, lags)

                    # lag_plot
                    x_site = min(ar_data['t'])
                    y_site = min(ar_data['t-1'])
                    self.__data_lag_R = round(Math.correlation_coefficient(ar_data['t'], ar_data['t-1']), 4)

                    # ACF_chart
                    acf = []
                    for i in range(1,lags+1):
                        acf_value = 0
                        acf_value = Math.ACF(ar_data['t'], ar_data['t-'+str(i)])
                        acf.append(acf_value)
                    index_list = Data.create_sequence_list(acf)
                    df_acf = Data.list_to_dataframe(acf, index_list)

                    # historocal_trend_line
                    ax1 = plt.subplot(G[0, :])
                    plt.title(" Historocal trend : " + self.__title_times_data, fontsize=self.__fontsize_title)
                    plt.xlabel(xlabel, fontsize = self.__fontsize_x_y_label)
                    plt.ylabel(ylabel, fontsize = self.__fontsize_x_y_label)
                    plt.tick_params(axis = 'both', labelsize = self.__fontsize_x_y, color = 'green')
                    plt.xticks(np.linspace(0, len(data), 12), rotation=15)
                    plt.plot(data)

                    # lag_plot
                    ax2 = plt.subplot(G[1, :2])
                    plt.title("Lag Plot : "+self.__title_times_data, fontsize=self.__fontsize_title)
                    plt.xlabel("y(t)", fontsize=self.__fontsize_x_y_label)
                    plt.ylabel("y(t-1)", fontsize=self.__fontsize_x_y_label)
                    plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
                    plt.text(x_site, y_site, r'$\mathrm{r} = $' + str(self.__data_lag_R),
                            bbox=dict(boxstyle="square",
                                                ec=(1., 0.5, 0.5),
                                                fc=(1., 0.9, 0.9),
                                                ), fontsize=self.__fontsize_x_y_label)
                    plt.scatter(ar_data['t'], ar_data['t-1'], s=10)

                    # ACF_chart
                    ax3 = plt.subplot(G[1, -2:])
                    plt.title("ACF : "+self.__title_times_data, fontsize=self.__fontsize_title)
                    plt.xlabel("lags", fontsize=self.__fontsize_x_y_label)
                    plt.ylabel("ACF", fontsize=self.__fontsize_x_y_label)
                    plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
                    plt.bar(df_acf.index, df_acf[0])
                    plt.tight_layout()

                    if not os.path.isdir(file_path + '/'):
                        os.mkdir(file_path)
                    plt.savefig(file_path + "/statistics_infographic_" + self.__title_times_data + ".png")
                    plt.show()
                    print('Saved successfully (statistics_infographic). File path = ' + file_path + "/statistics_infographic_" + self.__title_times_data + ".png")

                else:
                    raise ValueError("List content must be one numerical data: data=",chack_data)
            else:
                raise ValueError("'lags' value must be a positive integer, and the condition is greater than 1 and less than the length of the data.")
        except AttributeError as err:
            raise AttributeError(err)
        except TypeError as err:
            raise TypeError(err)

    def forecast_result_group_chart(self,train, test, model_1: Model, model_2: Model, file_path: str, model_1_name: str = "model1", model_2_name: str = "model2", xlabel: str = 'date', ylabel: str = 'price'):
        """Combine and compare the prediction results of the two models.

        The graph drawn by this function can visualize the data divided into the training set and the test set in the same historical data, 
        the prediction results of the test sets of the two models, and the model measurement indicators, such as MSE, MAE, RMSE, NRMSE, finally saved in the folder.

        Parameters
        ---------
            train: pandas.Series
            'train' is a one-dimensional numerical time series data of type'pandas.Series' taken from'pandas.DataFrame'.

            test: pandas.Series
            'test' is a one-dimensional numerical time series data of type'pandas.Series' taken from'pandas.DataFrame'.

            model_1: Model
            The object type is TimeSeriesAnalysis.Model. The First model.

            model_2: Model
            The object type is TimeSeriesAnalysis.Model. The second model.

            file_path: string
            Path to save image file.

            model_1_name: str, default  = "model1"
            The name of the model_1.

            model_2_name: str, default  = "model2"
            The name of the model_2.

            xlabel: str, default  = 'date'
            The name of the x-axis.

            ylabel: str, default  = 'price'
            The name of the y-axis

        Returns
        ---------
        No return value.

        Error
        ---------
            ValueError: List content must be one-dimensional numerical data: train=", chack_train, ", test=", chack_test).
            Solution: 'True': one-dimensional numerical list. For example:[1,5,8,6,3].    'False': two-dimensional lists, strings, and one-dimensional 
            non-numerical lists.For example: [[3, 6], [7, 2, 9, 5]] , '5' ,  [3, 6, 7, '2', 9, 5] . If'chack_train' is'True', it means that 'train' does not need to be 
            changed; if it is'False', the input'train' is changed to a one-dimensional list of numerical data.'test' has the same judgment and processing 
            method as 'train'.

            TypeError: object of type 'NoneType' has no len()
            Solution: 'train' or 'test' is None. Check that the input list is a one-dimensional list, for example: [1,5,8,6,3].

        """
        try:
            chack_train = Math.chack_list_all_num(train)
            chack_test = Math.chack_list_all_num(test)
            if(chack_train and chack_test):
                plt.figure(figsize=(15,9))
                G = gridspec.GridSpec(2, 4)

                # historocal_trend_line
                ax1 = plt.subplot(G[0, :])
                plt.title(" Historocal trend : " + self.__title_times_data, fontsize=self.__fontsize_title)
                plt.plot(train, color='lightgray', label = 'Train')
                plt.plot(test, color='gold', label = 'Test')
                plt.plot(model_1.test_predict, color='r', label=model_1_name)
                plt.plot(model_2.test_predict, label=model_2_name)
                plt.xlabel(xlabel, fontsize = self.__fontsize_x_y_label)
                plt.ylabel(ylabel, fontsize = self.__fontsize_x_y_label)
                plt.tick_params(axis = 'both', labelsize = self.__fontsize_x_y)
                plt.xticks(np.linspace(0, len(train)+ len(test), 12), rotation=15)
                ax1.legend()

                ax2 = plt.subplot(G[1, :2])
                plt.title("Forecast Result  " +  model_1_name + " : " + self.__title_times_data, fontsize=self.__fontsize_title)
                plt.plot(model_1.test_data, color='lightgray', label="Test")
                plt.plot(model_1.test_predict, color="r", label=model_1_name)
                plt.xlabel(xlabel, fontsize=self.__fontsize_x_y_label)
                plt.ylabel(ylabel, fontsize=self.__fontsize_x_y_label)
                x_site = len(model_1.test_data)
                if(min(model_1.test_data) < min(model_1.test_predict)):
                    y_site = min(model_1.test_data)
                else:
                    y_site = min(model_1.test_predict)
                round_mod_1_err = Math.round_list(model_1.test_error, 4)
                self.__model1_result = round_mod_1_err
                plt.text(x_site*0.05, y_site, r'$\mathrm{MSE} = $' + str(self.__model1_result[0]) + '\n' + r'$\mathrm{MAE} = $' + str(self.__model1_result[1]) +'\n'
                                                                    + r'$\mathrm{RMSE} = $' + str(self.__model1_result[2]) + '\n' + r'$\mathrm{NRMSE} = $' + str(self.__model1_result[3]),
                                    bbox=dict( boxstyle="square", ec=(1., 0.5, 0.5), fc=(1., 0.9, 0.9), ), fontsize=self.__fontsize_x_y)
                plt.xticks(np.linspace(0, len(model_1.test_data), 12), rotation=15)
                plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
                plt.legend(loc='upper left')

                ax3 = plt.subplot(G[1, -2:])
                plt.title("Forecast Result  " +  model_2_name + " : " + self.__title_times_data, fontsize=self.__fontsize_title)
                plt.plot(model_2.test_data, color='lightgray', label="Test")
                plt.plot(model_2.test_predict, label=model_2_name)
                plt.xlabel(xlabel, fontsize=self.__fontsize_x_y_label)
                plt.ylabel(ylabel, fontsize=self.__fontsize_x_y_label)
                x_site = len(model_2.test_data)
                if(min(model_2.test_data) < min(model_2.test_predict)):
                    y_site = min(model_2.test_data)
                else:
                    y_site = min(model_2.test_predict)
                round_mod_2_err = Math.round_list(model_2.test_error, 4)
                self.__model2_result = round_mod_2_err
                plt.text(x_site*0.05, y_site, r'$\mathrm{MSE} = $' + str(self.__model2_result[0]) + '\n' + r'$\mathrm{MAE} = $' + str(self.__model2_result[1]) +'\n'
                                                                    + r'$\mathrm{RMSE} = $' + str(self.__model2_result[2]) + '\n' + r'$\mathrm{NRMSE} = $' + str(self.__model2_result[3]),
                                    bbox=dict( boxstyle="square", ec=(1., 0.5, 0.5), fc=(1., 0.9, 0.9), ), fontsize=self.__fontsize_x_y)
                plt.xticks(np.linspace(0, len(model_2.test_data), 12), rotation=15)
                plt.tick_params(axis='both', labelsize=self.__fontsize_x_y)
                plt.legend(loc='upper left')
                plt.tight_layout()

                if not os.path.isdir(file_path + '/'):
                    os.mkdir(file_path)
                plt.savefig(file_path + "/forecast_result_group_chart_" + self.__title_times_data + ".png")
                plt.show()
                print('Saved successfully (forecast result group chart). File path = ' + file_path + "/forecast_result_group_chart_" + self.__title_times_data + ".png")
            else:
                raise ValueError("List content must be one-dimensional numerical data: train=", chack_train, ", test=", chack_test)

        except AttributeError as err:
            raise AttributeError(err)
        except TypeError as err:
            raise TypeError(err)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright 2020 by leodflag

"""
TimeSeriesAnalysis
=====

module
---------
Chart
    Time series data related drawing function.

Model
    This module contains time series models for forecasting.

ProcessData
    This module contains the access and processing methods of time series data.

"""



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="TimesML",
    version="0.0.7",
    author="leodflag",
    author_email="lovedoglion5@gmail.com",
    description="This package is used for time series data analysis. Correct internal errors of functions, such as Statistics.normalized_mean_squard_error, Statistics.coefficient_of_determination.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/leodflag/TimesML",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
    ],
    python_requires='>=3.6',
)
